import pytest
from datetime import datetime, UTC
from atkins.services.user_credit import UserCreditService


@pytest.fixture
def credit_service(db):
    service = UserCreditService(db)
    service.create_collections()
    service.build_index()
    return service


def test_get_credit_nonexistent(credit_service):
    user_id = "test_user"
    credit = credit_service.get_credit(user_id)
    assert credit is None


def test_add_credit(credit_service):
    user_id = "test_user"
    amount = 100
    reason = {"type": "test", "description": "test credit"}
    
    # Add credit
    assert credit_service.add_credit(user_id, amount, reason) is True
    
    # Verify credit
    credit = credit_service.get_credit(user_id)
    assert credit is not None
    assert credit["user"] == user_id
    assert credit["credit"] == amount
    assert len(credit["history"]) == 1
    assert credit["history"][0]["amount"] == amount
    assert credit["history"][0]["reason"] == reason
    assert isinstance(credit["history"][0]["time"], datetime)


def test_minus_credit(credit_service):
    user_id = "test_user"
    initial_amount = 100
    minus_amount = 50
    reason = {"type": "test", "description": "test credit"}
    
    # Add initial credit
    credit_service.add_credit(user_id, initial_amount, reason)
    
    # Minus credit
    assert credit_service.minus_credit(user_id, minus_amount, reason) is True
    
    # Verify credit
    credit = credit_service.get_credit(user_id)
    assert credit is not None
    assert credit["credit"] == initial_amount - minus_amount
    assert len(credit["history"]) == 2
    assert credit["history"][1]["amount"] == -minus_amount


def test_credit_history(credit_service):
    user_id = "test_user"
    reason = {"type": "test", "description": "test credit"}
    
    # Add multiple credits
    credit_service.add_credit(user_id, 100, reason)
    credit_service.add_credit(user_id, 50, reason)
    credit_service.minus_credit(user_id, 30, reason)
    
    # Verify history
    credit = credit_service.get_credit(user_id)
    assert credit is not None
    assert credit["credit"] == 120  # 100 + 50 - 30
    assert len(credit["history"]) == 3
    assert credit["history"][0]["amount"] == 100
    assert credit["history"][1]["amount"] == 50
    assert credit["history"][2]["amount"] == -30 


import pytest
from datetime import datetime, UTC
from atkins.services.redeem_code import RedeemCodeService, insert_dashes


@pytest.fixture
def redeem_code_service(db):
    service = RedeemCodeService(db)
    service.create_collections()
    service.build_index()
    return service


def test_insert_dashes():
    assert insert_dashes("12345678") == "1234-5678"
    assert insert_dashes("123456789") == "1234-5678-9"
    assert insert_dashes("") == ""


def test_generate_secure_code():
    code = RedeemCodeService.generate_secure_code()
    assert len(code.replace("-", "")) == 32  # 16 bytes = 32 hex chars
    assert "-" in code
    assert len(code.split("-")) == 8  # 32 chars / 4 = 8 segments


def test_create_code(redeem_code_service):
    code = "TEST-1234-5678"
    code_type = "premium_membership"
    
    # Create code
    created_code = redeem_code_service.create_code(code, code_type)
    assert created_code == code
    
    # Verify code exists
    code_record = redeem_code_service.get_code(code)
    assert code_record is not None
    assert code_record["code"] == code
    assert code_record["code_type"] == code_type
    assert code_record["used"] is False
    assert code_record["status"] == "issued"
    assert isinstance(code_record["createdAt"], datetime)
    assert code_record["request"] is None
    assert code_record["response"] is None
    assert code_record["error"] is None


def test_create_codes(redeem_code_service):
    codes = ["TEST-1234-5678", "TEST-8765-4321", "TEST-1111-2222"]
    code_type = "premium_membership"
    
    # Create multiple codes
    created_docs = redeem_code_service.create_codes(codes, code_type)
    
    # Verify all codes were created
    assert len(created_docs) == len(codes)
    
    # Verify each code exists and has correct fields
    for code in codes:
        code_record = redeem_code_service.get_code(code)
        assert code_record is not None
        assert code_record["code"] == code
        assert code_record["code_type"] == code_type
        assert code_record["used"] is False
        assert code_record["status"] == "issued"
        assert isinstance(code_record["createdAt"], datetime)
        assert code_record["request"] is None
        assert code_record["response"] is None
        assert code_record["error"] is None


def test_create_codes_duplicate(redeem_code_service):
    codes = ["TEST-1234-5678", "TEST-1234-5678"]  # Duplicate code
    code_type = "premium_membership"
    
    # Try to create codes with duplicates
    with pytest.raises(Exception):  # Should raise duplicate key error
        redeem_code_service.create_codes(codes, code_type)


def test_create_codes_empty_list(redeem_code_service):
    code_type = "premium_membership"
    # Create empty list of codes
    created_docs = redeem_code_service.create_codes([], code_type)
    assert len(created_docs) == 0


def test_start_processing_code(redeem_code_service):
    code = "TEST-1234-5678"
    code_type = "premium_membership"
    request = {"type": "test", "value": 100}
    
    # Create code
    redeem_code_service.create_code(code, code_type)
    
    # Start processing
    assert redeem_code_service.start_processing_code(code, request) is True
    
    # Verify processing state
    code_record = redeem_code_service.get_code(code)
    assert code_record["status"] == "processing"
    assert code_record["request"] == request
    assert isinstance(code_record["updatedAt"], datetime)


def test_mark_processing_failed(redeem_code_service):
    code = "TEST-1234-5678"
    code_type = "premium_membership"
    error_msg = {"message": "Test error", "code": "TEST_ERROR"}
    
    # Create code and start processing
    redeem_code_service.create_code(code, code_type)
    redeem_code_service.start_processing_code(code, {"type": "test"})
    
    # Mark as failed
    assert redeem_code_service.mark_processing_failed(code, error_msg) is True
    
    # Verify failed state
    code_record = redeem_code_service.get_code(code)
    assert code_record["status"] == "error"
    assert code_record["error"] == error_msg
    assert isinstance(code_record["updatedAt"], datetime)


def test_mark_processing_done(redeem_code_service):
    code = "TEST-1234-5678"
    code_type = "premium_membership"
    response = {"success": True, "message": "Test completed"}
    
    # Create code and start processing
    redeem_code_service.create_code(code, code_type)
    redeem_code_service.start_processing_code(code, {"type": "test"})
    
    # Mark as done
    assert redeem_code_service.mark_processing_done(code, response) is True
    
    # Verify completed state
    code_record = redeem_code_service.get_code(code)
    assert code_record["status"] == "used"
    assert code_record["response"] == response
    assert isinstance(code_record["updatedAt"], datetime)


def test_get_nonexistent_code(redeem_code_service):
    code = "NONEXISTENT-CODE"
    code_record = redeem_code_service.get_code(code)
    assert code_record is None


def test_unique_code_constraint(redeem_code_service):
    code = "TEST-1234-5678"
    code_type = "premium_membership"
    
    # Create code
    redeem_code_service.create_code(code, code_type)
    
    # Try to create same code again
    with pytest.raises(Exception):  # Should raise duplicate key error
        redeem_code_service.create_code(code, code_type) 


from datetime import datetime
from secrets import token_hex

from pymongo.synchronous.collection import Collection

from atkins.base import MongoService
from typing import TypedDict, Optional, Literal


class RedeemCodeRecord(TypedDict):
    code: str
    code_type: str  # Indicates which product this code can redeem

    used: bool
    createdAt: datetime
    updatedAt: Optional[datetime]

    status: Literal['issued', 'used', 'expired', 'processing', 'error']

    request: Optional[dict]
    response: Optional[dict]
    error: Optional[dict]


def insert_dashes(input_string: str) -> str:
    """
    Inserts a dash every 4 characters in the input string.

    :param input_string: The string to process
    :return: A new string with dashes inserted
    """
    return '-'.join(input_string[i:i + 4] for i in range(0, len(input_string), 4))


class RedeemCodeService(MongoService):
    def __init__(
            self,
            db,
            collection_name: str = 'redeem_codes'
    ):
        super().__init__(db)
        self.collection_name = collection_name
        self.coll: Collection[RedeemCodeRecord] = self.db[collection_name]

    def create_collections(
            self,
    ):
        self.db.create_collection(
            self.collection_name,
        )

    def build_index(self, **kwargs):
        """
        create the following indices:
         1. unique index on code field
         2. non-unique index on createdAt field
         2. non-unique index on status field


        :param kwargs:
        :return:
        """
        self.coll = self.db[self.collection_name]
        self.coll.create_index("code", unique=True)
        self.coll.create_index("createdAt")
        self.coll.create_index("status")

    @staticmethod
    def generate_secure_code(length=16):
        h = token_hex(length)
        return insert_dashes(h)

    def create_codes(self, codes: list[str], code_type: str) -> str:
        """
        Create a list of redeem code
        :param codes: list of code string
        :param code_type: type of the code, indicating which product it can redeem
        :return: The created code
        """

        if len(codes) == 0:
            return []

        docs = [
            {
                "code": code,
                "code_type": code_type,
                "used": False,
                "createdAt": datetime.now(),
                "updatedAt": None,
                "status": "issued",
                "request": None,
                "response": None,
                "error": None,
            } for code in codes
        ]
        self.coll.insert_many(docs)
        return docs

    def create_code(self, code: str, code_type: str) -> str:
        """
        Create a new redeem code
        :param code: The code string
        :param code_type: type of the code, indicating which product it can redeem
        :return: The created code
        """
        doc = {
            "code": code,
            "code_type": code_type,
            "used": False,
            "createdAt": datetime.now(),
            "updatedAt": None,
            "status": "issued",
            "request": None,
            "response": None,
            "error": None,
        }
        self.coll.insert_one(doc)
        return code

    def start_processing_code(self, code: str, request: dict) -> bool:
        update_data = {
            'status': 'processing',
            'request': request,
            'updatedAt': datetime.now(),
        }
        result = self.coll.update_one(
            {"code": code},
            {"$set": update_data}
        )
        return result.modified_count > 0

    def mark_processing_failed(self, code: str, error_msg: dict) -> bool:
        update_data = {
            'status': 'error',
            "error": error_msg,
            'updatedAt': datetime.now(),
        }
        result = self.coll.update_one(
            {"code": code},
            {"$set": update_data}
        )
        return result.modified_count > 0

    def mark_processing_done(self, code: str, response: dict) -> bool:
        update_data = {
            'status': 'used',
            "response": response,
            'updatedAt': datetime.now(),
        }
        result = self.coll.update_one(
            {"code": code},
            {"$set": update_data}
        )
        return result.modified_count > 0

    def get_code(self, code: str) -> Optional[RedeemCodeRecord]:
        """
        Get a code record by code
        :param code: The code to look up
        :return: The code record or None if not found
        """
        return self.coll.find_one({"code": code})



from datetime import timedelta


def get_timeduration(duration_str: str) -> timedelta:
    """
    Parse duration string to a timedelta object.
    Supported formats: 1m (minute), 2h (hour), 3d (day), 4w (week), 5y (year)
    :param duration_str: Duration string (e.g. "1m", "2h", "3d", "4w", "5y")
    :return: A timedelta object representing the parsed duration
    :raises ValueError: If the input string format is invalid
    """
    if not duration_str or not isinstance(duration_str, str):
        raise ValueError("Invalid duration string provided.")

    # Define multipliers for each unit
    multipliers = {
        "m": timedelta(minutes=1),
        "h": timedelta(hours=1),
        "d": timedelta(days=1),
        "w": timedelta(weeks=1),
        "y": timedelta(days=365)  # Approximate a year as 365 days
    }

    # Extract the numeric value and unit
    try:
        value = int(duration_str[:-1])  # Strip the last character (unit) and convert to integer
        unit = duration_str[-1]  # Extract the unit
    except (ValueError, IndexError):
        raise ValueError("Duration string must be in the format '<number><unit>' (e.g., '1m', '2h', '3d').")

    # Calculate timedelta based on unit
    if unit in multipliers:
        return value * multipliers[unit]
    else:
        raise ValueError(
            f"Unsupported duration unit '{unit}'. Supported units: 'm' (minute), 'h' (hour), 'd' (day), 'w' (week), 'y' (year).")



import pytest
from datetime import datetime, UTC, timedelta
from atkins.services.user_membership import UserMembershipService


@pytest.fixture
def membership_service(db):
    service = UserMembershipService(db)
    service.create_collections()
    service.build_index()
    return service


def test_list_membership_nonexistent(membership_service):
    user_id = "test_user"
    membership = membership_service.list_membership(user_id)
    assert membership is None


def test_assign_membership(membership_service):
    user_id = "test_user"
    membership_type = "premium"
    reason = {"type": "test", "description": "test membership"}
    
    # Assign membership
    assert membership_service.assign_membership(user_id, membership_type, reason) is True
    
    # Verify membership
    membership = membership_service.list_membership(user_id)
    assert membership is not None
    assert membership["user"] == user_id
    assert membership["membership"] == membership_type
    assert membership["auto_renew"] is True
    assert isinstance(membership["updatedAt"], datetime)
    assert isinstance(membership["validUntil"], datetime)
    assert len(membership["history"]) == 1
    assert membership["history"][0]["action"] == "assign"
    assert membership["history"][0]["reason"] == reason


def test_cancel_membership(membership_service):
    user_id = "test_user"
    membership_type = "premium"
    reason = {"type": "test", "description": "test membership"}
    cancel_reason = {"type": "test", "description": "cancel membership"}
    
    # Assign membership
    membership_service.assign_membership(user_id, membership_type, reason)
    
    # Cancel membership
    assert membership_service.cancel_membership(user_id, membership_type, cancel_reason) is True
    
    # Verify membership
    membership = membership_service.list_membership(user_id)
    assert membership is not None
    assert membership["auto_renew"] is False
    assert len(membership["history"]) == 2
    assert membership["history"][1]["action"] == "cancel"
    assert membership["history"][1]["reason"] == cancel_reason


def test_set_auto_renew_preference(membership_service):
    user_id = "test_user"
    membership_type = "premium"
    reason = {"type": "test", "description": "test membership"}
    
    # Assign membership
    membership_service.assign_membership(user_id, membership_type, reason)
    
    # Set auto-renew preference
    assert membership_service.set_membership_auto_renew_preference(user_id, membership_type, False) is True
    
    # Verify preference
    membership = membership_service.list_membership(user_id)
    assert membership["auto_renew"] is False
    
    # Change preference back
    assert membership_service.set_membership_auto_renew_preference(user_id, membership_type, True) is True
    
    # Verify preference
    membership = membership_service.list_membership(user_id)
    assert membership["auto_renew"] is True


def test_membership_history(membership_service):
    user_id = "test_user"
    membership_type = "premium"
    reason = {"type": "test", "description": "test membership"}
    cancel_reason = {"type": "test", "description": "cancel membership"}
    
    # Assign membership
    membership_service.assign_membership(user_id, membership_type, reason)
    
    # Cancel membership
    membership_service.cancel_membership(user_id, membership_type, cancel_reason)
    
    # Verify history
    membership = membership_service.list_membership(user_id)
    assert len(membership["history"]) == 2
    assert membership["history"][0]["action"] == "assign"
    assert membership["history"][1]["action"] == "cancel" 


import os.path

import pytest
from pymongo import MongoClient
from docker import DockerClient
from docker.errors import NotFound
import time


@pytest.fixture(scope="session")
def docker_client():
    sock_path = os.path.expanduser('~/.docker/run/docker.sock')
    sock_url = f'unix:/{sock_path}'
    return DockerClient(base_url=sock_url)


@pytest.fixture(scope="session")
def mongodb_container(docker_client):
    container_name = "atkins-test-mongodb"

    try:
        existing_container = docker_client.containers.get(container_name)
        if existing_container.status == "running":
            pass
    except NotFound:

        # Start MongoDB container
        container = docker_client.containers.run(
            "mongo:latest",
            name="atkins-test-mongodb",
            detach=True,
            ports={'27017/tcp': ('127.0.0.1', 47017)},
            remove=True
        )

        time.sleep(2)

    yield container

    # Cleanup
    try:
        container.stop()
    except NotFound:
        pass


@pytest.fixture
def mongodb_client(mongodb_container):
    client = MongoClient('mongodb://localhost:47017')
    yield client
    # Drop all databases between tests
    for db_name in client.list_database_names():
        if db_name not in ['admin', 'local', 'config']:
            client.drop_database(db_name)
    client.close()


@pytest.fixture
def db(mongodb_client):
    return mongodb_client['test_db']



from datetime import datetime

import pytest

from atkins.services.activity_logging import ActivityLoggerService


@pytest.fixture
def activity_logger(db):
    service = ActivityLoggerService(
        db,
        use_ttl=False,  # Changed to False since we're using capped collection
        use_capped_coll=True,
        capped_coll_size=1024 * 1024 * 1024,  # 1GB
        capped_coll_records=1000,
        collection_name='test_activity_logs'
    )
    service.create_collections()
    service.build_index()
    return service


@pytest.fixture
def ttl_logger(db):
    service = ActivityLoggerService(
        db,
        use_ttl=True,
        use_capped_coll=False,
        ttl_limit=60 * 60 * 24 * 10,  # 10 days
        collection_name='test_ttl_logs'
    )
    service.create_collections()
    service.build_index()
    return service


def test_ttl_and_capped_conflict(db):
    with pytest.raises(ValueError, match="TTL and Capped collections cannot be used at the same time."):
        ActivityLoggerService(
            db,
            use_ttl=True,
            use_capped_coll=True,
            collection_name='test_conflict_logs'
        )


def test_log_activity(activity_logger):
    user = "test_user"
    target = "test_target"
    meta = {"action": "test", "value": 123}
    
    # Log activity
    activity_logger.log_activity(user, target, meta)
    
    # Verify log exists
    logs = list(activity_logger.fetch_logs())
    assert len(logs) == 1
    log = logs[0]
    assert log["user"] == user
    assert log["target"] == target
    assert log["action"] == meta["action"]
    assert log["value"] == meta["value"]
    assert isinstance(log["createdAt"], datetime)


def test_log_activity_no_meta(activity_logger):
    user = "test_user"
    target = "test_target"
    
    # Log activity without meta
    activity_logger.log_activity(user, target)
    
    # Verify log exists
    logs = list(activity_logger.fetch_logs())
    assert len(logs) == 1
    log = logs[0]
    assert log["user"] == user
    assert log["target"] == target
    assert "action" not in log
    assert isinstance(log["createdAt"], datetime)


def test_fetch_logs_with_period(activity_logger):
    user = "test_user"
    target = "test_target"
    
    # Log multiple activities
    for i in range(3):
        activity_logger.log_activity(user, target, {"index": i})
    
    # Fetch logs with period
    logs = list(activity_logger.fetch_logs(period="1d"))
    assert len(logs) == 3
    
    # Verify logs are in chronological order
    for i in range(len(logs) - 1):
        assert logs[i]["createdAt"] <= logs[i + 1]["createdAt"]


def test_fetch_logs_with_end_date(activity_logger):
    user = "test_user"
    target = "test_target"
    end_date = datetime.now()
    
    # Log activity
    activity_logger.log_activity(user, target)
    
    # Fetch logs with end date
    logs = list(activity_logger.fetch_logs(end=end_date))
    assert len(logs) == 1
    assert logs[0]["createdAt"] <= end_date


def test_capped_collection(activity_logger):
    user = "test_user"
    target = "test_target"
    
    # Log more activities than the capped collection size
    for i in range(2000):  # More than the 1000 record limit
        activity_logger.log_activity(user, target, {"index": i})
    
    # Verify collection is capped
    logs = list(activity_logger.fetch_logs())
    assert len(logs) <= 1000  # Should not exceed the capped size


def test_ttl_index(ttl_logger):
    user = "test_user"
    target = "test_target"
    
    # Log activity
    ttl_logger.log_activity(user, target)
    
    # Verify index has expireAfterSeconds set
    indexes = ttl_logger.coll.index_information()
    assert "createdAt_1" in indexes
    assert indexes["createdAt_1"].get("expireAfterSeconds") == 60 * 60 * 24 * 10  # 10 days


def test_custom_ttl_limit(db):
    custom_ttl = 60 * 60 * 24 * 5  # 5 days
    service = ActivityLoggerService(
        db,
        use_ttl=True,
        use_capped_coll=False,
        ttl_limit=custom_ttl,
        collection_name='test_custom_ttl_logs'
    )
    service.create_collections()
    service.build_index()
    
    # Verify custom TTL is set
    indexes = service.coll.index_information()
    assert "createdAt_1" in indexes
    assert indexes["createdAt_1"].get("expireAfterSeconds") == custom_ttl



from datetime import datetime

from pymongo.synchronous.collection import Collection

from atkins.base import MongoService
from typing import TypedDict, Optional, Literal

from atkins.utils import get_timeduration


class MembershipUpdateRecord(TypedDict):
    time: datetime
    reason: dict
    action: Literal['assign', 'cancel', 'renew']


class UserMembershipRecord(TypedDict):
    user: str
    membership: str
    updatedAt: datetime
    validUntil: datetime
    history: list[MembershipUpdateRecord]
    auto_renew: bool


class UserMembershipService(MongoService):
    def __init__(
            self,
            db,
            collection_name: str = 'user_membership'
    ):
        super().__init__(db)
        self.collection_name = collection_name
        self.coll: Collection[UserMembershipRecord] = self.db[collection_name]

    def create_collections(
            self,
    ):
        self.db.create_collection(
            self.collection_name,
        )

    def build_index(self, **kwargs):
        """
        create the following indices:
         1. unique index on user field
         2. non-unique index on membership, validUntil field
        """
        self.coll = self.db[self.collection_name]
        self.coll.create_index("user", unique=True)
        self.coll.create_index([("membership", 1), ("validUntil", 1)])

    def list_membership(self, user: str) -> Optional[UserMembershipRecord]:
        """
        Get user's membership record
        :param user: User identifier
        :return: User's membership record or None if not found
        """
        return self.coll.find_one({"user": user})

    def assign_membership(self, user: str, membership: str, reason: dict, duration: str = "1y") -> bool:
        """
        Assign membership to user
        :param user: User identifier
        :param membership: Membership type
        :param reason: Reason for assigning membership
        :param duration: Duration string (e.g. "1m", "2h", "3d", "4w", "5y")
        :return: True if update was successful
        """
        now = datetime.now()
        valid_until = now + get_timeduration(duration)

        update_record = MembershipUpdateRecord(
            time=now,
            reason=reason,
            action="assign"
        )

        result = self.coll.update_one(
            {"user": user},
            {
                "$set": {
                    "membership": membership,
                    "updatedAt": now,
                    "validUntil": valid_until,
                    "auto_renew": True
                },
                "$push": {"history": update_record}
            },
            upsert=True
        )
        return result.modified_count > 0 or result.upserted_id is not None

    def cancel_membership(self, user: str, membership: str, reason: dict) -> bool:
        """
        Cancel user's membership
        :param user: User identifier
        :param membership: Membership type to cancel
        :param reason: Reason for cancellation
        :return: True if update was successful
        """
        now = datetime.now()
        update_record = MembershipUpdateRecord(
            time=now,
            reason=reason,
            action="cancel"
        )

        result = self.coll.update_one(
            {"user": user, "membership": membership},
            {
                "$set": {
                    "validUntil": now,
                    "auto_renew": False,
                    "updatedAt": now
                },
                "$push": {"history": update_record}
            }
        )
        return result.modified_count > 0

    def set_membership_auto_renew_preference(self, user: str, membership: str, auto_renew_preference: bool) -> bool:
        """
        Set user's membership auto-renew preference
        :param user: User identifier
        :param membership: Membership type
        :param auto_renew_preference: Whether to auto-renew
        :return: True if update was successful
        """
        result = self.coll.update_one(
            {"user": user, "membership": membership},
            {"$set": {"auto_renew": auto_renew_preference}}
        )
        return result.modified_count > 0



"""
This implementation is from Werkzeug.

https://github.com/pallets/werkzeug/blob/main/src/werkzeug/security.py

Copyright 2007 Pallets

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

3.  Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""




from __future__ import annotations

import hashlib
import hmac
import os
import posixpath
import secrets

SALT_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
DEFAULT_PBKDF2_ITERATIONS = 1_000_000

_os_alt_seps: list[str] = list(
    sep for sep in [os.sep, os.path.altsep] if sep is not None and sep != "/"
)


def gen_salt(length: int) -> str:
    """Generate a random string of SALT_CHARS with specified ``length``."""
    if length <= 0:
        raise ValueError("Salt length must be at least 1.")

    return "".join(secrets.choice(SALT_CHARS) for _ in range(length))


def _hash_internal(method: str, salt: str, password: str) -> tuple[str, str]:
    method, *args = method.split(":")
    salt_bytes = salt.encode()
    password_bytes = password.encode()

    if method == "scrypt":
        if not args:
            n = 2**15
            r = 8
            p = 1
        else:
            try:
                n, r, p = map(int, args)
            except ValueError:
                raise ValueError("'scrypt' takes 3 arguments.") from None

        maxmem = 132 * n * r * p  # ideally 128, but some extra seems needed
        return (
            hashlib.scrypt(
                password_bytes, salt=salt_bytes, n=n, r=r, p=p, maxmem=maxmem
            ).hex(),
            f"scrypt:{n}:{r}:{p}",
        )
    elif method == "pbkdf2":
        len_args = len(args)

        if len_args == 0:
            hash_name = "sha256"
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 1:
            hash_name = args[0]
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 2:
            hash_name = args[0]
            iterations = int(args[1])
        else:
            raise ValueError("'pbkdf2' takes 2 arguments.")

        return (
            hashlib.pbkdf2_hmac(
                hash_name, password_bytes, salt_bytes, iterations
            ).hex(),
            f"pbkdf2:{hash_name}:{iterations}",
        )
    else:
        raise ValueError(f"Invalid hash method '{method}'.")


def generate_password_hash(
    password: str, method: str = "scrypt", salt_length: int = 16
) -> str:
    """Securely hash a password for storage. A password can be compared to a stored hash
    using :func:`check_password_hash`.

    The following methods are supported:

    -   ``scrypt``, the default. The parameters are ``n``, ``r``, and ``p``, the default
        is ``scrypt:32768:8:1``. See :func:`hashlib.scrypt`.
    -   ``pbkdf2``, less secure. The parameters are ``hash_method`` and ``iterations``,
        the default is ``pbkdf2:sha256:600000``. See :func:`hashlib.pbkdf2_hmac`.

    Default parameters may be updated to reflect current guidelines, and methods may be
    deprecated and removed if they are no longer considered secure. To migrate old
    hashes, you may generate a new hash when checking an old hash, or you may contact
    users with a link to reset their password.

    :param password: The plaintext password.
    :param method: The key derivation function and parameters.
    :param salt_length: The number of characters to generate for the salt.

    .. versionchanged:: 3.1
        The default iterations for pbkdf2 was increased to 1,000,000.

    .. versionchanged:: 2.3
        Scrypt support was added.

    .. versionchanged:: 2.3
        The default iterations for pbkdf2 was increased to 600,000.

    .. versionchanged:: 2.3
        All plain hashes are deprecated and will not be supported in Werkzeug 3.0.
    """
    salt = gen_salt(salt_length)
    h, actual_method = _hash_internal(method, salt, password)
    return f"{actual_method}${salt}${h}"


def check_password_hash(pwhash: str, password: str) -> bool:
    """Securely check that the given stored password hash, previously generated using
    :func:`generate_password_hash`, matches the given password.

    Methods may be deprecated and removed if they are no longer considered secure. To
    migrate old hashes, you may generate a new hash when checking an old hash, or you
    may contact users with a link to reset their password.

    :param pwhash: The hashed password.
    :param password: The plaintext password.

    .. versionchanged:: 2.3
        All plain hashes are deprecated and will not be supported in Werkzeug 3.0.
    """
    try:
        method, salt, hashval = pwhash.split("$", 2)
    except ValueError:
        return False

    return hmac.compare_digest(_hash_internal(method, salt, password)[0], hashval)




import abc
from pymongo.synchronous.database import Database


class MongoService:
    db: Database

    def __init__(self, db: Database) -> None:
        self.db = db

    @abc.abstractmethod
    def build_index(self, **kwargs):
        pass

    @abc.abstractmethod
    def create_collections(self, **kwargs):
        pass



from datetime import datetime

from pymongo.synchronous.collection import Collection

from atkins.base import MongoService
from typing import TypedDict, Optional, Literal


class CreditUpdateRecord(TypedDict):
    time: datetime
    reason: dict
    amount: int


class UserCreditRecord(TypedDict):
    user: str
    credit: int
    updatedAt: datetime
    history: list[CreditUpdateRecord]


class UserCreditService(MongoService):
    def __init__(
            self,
            db,
            collection_name: str = 'user_credit'
    ):
        super().__init__(db)
        self.collection_name = collection_name
        self.coll: Collection[UserCreditRecord] = self.db[collection_name]

    def create_collections(
            self,
    ):
        self.db.create_collection(
            self.collection_name,
        )

    def build_index(self, **kwargs):
        """
        create the following indices:
         1. unique index on user field
        """
        self.coll = self.db[self.collection_name]
        self.coll.create_index("user", unique=True)

    def get_credit(self, user: str) -> Optional[UserCreditRecord]:
        """
        Get user's credit record
        :param user: User identifier
        :return: User's credit record or None if not found
        """
        return self.coll.find_one({"user": user})

    def add_credit(self, user: str, amount: int, reason: dict) -> bool:
        """
        Add credit to user's account
        :param user: User identifier
        :param amount: Amount of credit to add
        :param reason: Reason for adding credit
        :return: True if update was successful
        """
        now = datetime.now()
        update_record = CreditUpdateRecord(
            time=now,
            reason=reason,
            amount=amount
        )

        result = self.coll.update_one(
            {"user": user},
            {
                "$inc": {"credit": amount},
                "$set": {"updatedAt": now},
                "$push": {"history": update_record}
            },
            upsert=True
        )
        return result.modified_count > 0 or result.upserted_id is not None

    def minus_credit(self, user: str, amount: int, reason: dict) -> bool:
        """
        Subtract credit from user's account
        :param user: User identifier
        :param amount: Amount of credit to subtract
        :param reason: Reason for subtracting credit
        :return: True if update was successful
        """
        now = datetime.now()
        update_record = CreditUpdateRecord(
            time=now,
            reason=reason,
            amount=-amount
        )

        result = self.coll.update_one(
            {"user": user},
            {
                "$inc": {"credit": -amount},
                "$set": {"updatedAt": now},
                "$push": {"history": update_record}
            },
            upsert=True
        )
        return result.modified_count > 0 or result.upserted_id is not None



import pytest
from datetime import datetime, UTC
from atkins.services.user import UserService


@pytest.fixture
def user_service(db):
    service = UserService(db)
    service.create_collections()
    service.build_index()
    return service


def test_create_user(user_service):
    user_id = "test_user"
    email = "test@example.com"
    password = "password123"
    
    # Create user
    created_user = user_service.create_user(user_id, email, password)
    assert created_user == user_id
    
    # Verify user exists
    user = user_service.get_user(user_id)
    assert user is not None
    assert user["user"] == user_id
    assert user["email"] == email
    assert user["verified"] is False
    assert isinstance(user["createdAt"], datetime)


def test_verify_user_password(user_service):
    user_id = "test_user"
    email = "test@example.com"
    password = "password123"
    
    # Create user
    user_service.create_user(user_id, email, password)
    
    # Test correct password
    assert user_service.verify_user_password(user_id, password) is True
    
    # Test incorrect password
    assert user_service.verify_user_password(user_id, "wrong_password") is False


def test_update_user_password(user_service):
    user_id = "test_user"
    email = "test@example.com"
    password = "password123"
    new_password = "new_password123"
    
    # Create user
    user_service.create_user(user_id, email, password)
    
    # Update password
    assert user_service.update_user_password(user_id, new_password) is True
    
    # Verify old password doesn't work
    assert user_service.verify_user_password(user_id, password) is False
    
    # Verify new password works
    assert user_service.verify_user_password(user_id, new_password) is True


def test_change_user_email(user_service):
    user_id = "test_user"
    email = "test@example.com"
    new_email = "new@example.com"
    password = "password123"
    
    # Create user
    user_service.create_user(user_id, email, password)
    
    # Change email
    assert user_service.change_user_email(user_id, new_email) is True
    
    # Verify email changed
    user = user_service.get_user(user_id)
    assert user["email"] == new_email


def test_change_user_verification_status(user_service):
    user_id = "test_user"
    email = "test@example.com"
    password = "password123"
    
    # Create user
    user_service.create_user(user_id, email, password)
    
    # Verify initial status
    user = user_service.get_user(user_id)
    assert user["verified"] is False
    
    # Change verification status
    assert user_service.change_user_verification_status(user_id, True) is True
    
    # Verify status changed
    user = user_service.get_user(user_id)
    assert user["verified"] is True 


from datetime import datetime

from pymongo.synchronous.collection import Collection

from atkins.base import MongoService
from typing import TypedDict, Optional

from atkins.security import generate_password_hash, check_password_hash


class UserRecord(TypedDict):
    user: str
    email: str
    hashed_password: str
    verified: bool
    createdAt: datetime


class UserService(MongoService):

    def __init__(
            self,
            db,
            collection_name: str = 'users'
    ):
        super().__init__(db)
        self.collection_name = collection_name
        self.coll: Collection[UserRecord] = self.db[collection_name]

    def create_collections(
            self,
    ):
        self.db.create_collection(
            self.collection_name,
        )

    def build_index(self, **kwargs):
        """
        create the following indices:
         1. unique index on user field
         2. non-unique index on email field
         3. non-unique index on verified field
         4. non-unique index on createdAt field
        """
        self.coll = self.db[self.collection_name]
        self.coll.create_index("user", unique=True)
        self.coll.create_index("email")
        self.coll.create_index("verified")
        self.coll.create_index("createdAt")

    def create_user(self, user: str, email: str, password: str, verified: bool = False) -> str:
        """
        Create a new user
        :param user: User identifier
        :param email: User's email
        :param password: User's password
        :param verified: Whether the user is verified
        :return: Created user identifier
        """
        doc = {
            "user": user,
            "email": email,
            "hashed_password": generate_password_hash(password),
            "verified": verified,
            "createdAt": datetime.now()
        }
        self.coll.insert_one(doc)
        return user

    def update_user_password(self, user: str, password: str) -> bool:
        """
        Verify user's password
        :param user: User identifier
        :param password: new password
        :return: True if password matches
        """
        result = self.coll.update_one(
            {"user": user},
            {"$set": {"hashed_password": generate_password_hash(password)}}
        )
        return result.modified_count > 0

    def verify_user_password(self, user: str, password: str) -> bool:
        """
        Verify user's password
        :param user: User identifier
        :param password: Password to verify
        :return: True if password matches
        """
        user_record = self.coll.find_one({"user": user})
        if not user_record:
            return False
        return check_password_hash(user_record["hashed_password"], password)

    def change_user_email(self, user: str, email: str) -> bool:
        """
        Change user's email
        :param user: User identifier
        :param email: New email
        :return: True if update was successful
        """
        result = self.coll.update_one(
            {"user": user},
            {"$set": {"email": email}}
        )
        return result.modified_count > 0

    def change_user_verification_status(self, user: str, verification_status: bool) -> bool:
        """
        Change user's verification status
        :param user: User identifier
        :param verification_status: New verification status
        :return: True if update was successful
        """
        result = self.coll.update_one(
            {"user": user},
            {"$set": {"verified": verification_status}}
        )
        return result.modified_count > 0

    def get_user(self, user: str) -> Optional[UserRecord]:
        """
        Get user record
        :param user: User identifier
        :return: User record or None if not found
        """
        return self.coll.find_one({"user": user})



from datetime import datetime

from atkins.base import MongoService
from atkins.utils import get_timeduration


class ActivityLoggerService(MongoService):

    def __init__(
            self,
            db,
            use_ttl=True,
            use_capped_coll=True,
            capped_coll_size=1024 * 1024 * 1024,  # 1GB,
            capped_coll_records=1000,
            ttl_limit=60 * 60 * 24 * 10,  # 10 days.
            collection_name: str = 'activity_logs'
    ):
        super().__init__(db)
        self.collection_name = collection_name
        self.coll = self.db[collection_name]
        self.use_ttl = use_ttl
        self.use_capped_coll = use_capped_coll
        self.capped_coll_size = capped_coll_size
        self.capped_coll_records = capped_coll_records
        self.ttl_limit = ttl_limit

        if self.use_ttl and self.use_capped_coll:
            raise ValueError("TTL and Capped collections cannot be used at the same time.")

    def create_collections(
            self,
    ):

        args = {

        }

        if self.use_capped_coll:
            args["capped"] = True
            args["size"] = self.capped_coll_size
            args["max"] = self.capped_coll_records

        self.db.create_collection(
            self.collection_name,
            **args
        )
        self.coll = self.db[self.collection_name]

    def build_index(self, **kwargs):
        args = {}
        if self.use_ttl:
            args['expireAfterSeconds'] = self.ttl_limit
        self.coll.create_index(
            "createdAt",

            **args
        )

    def log_activity(self, user: str, target: str, meta=None):
        doc = {
            "user": user,
            "target": target,
            "createdAt": datetime.now()
        }

        if meta:
            doc.update(meta)

        self.coll.insert_one(doc)

    def fetch_logs(self, end=None, period='10d'):
        if end is None:
            end = datetime.now()
        start = end - get_timeduration(period)
        return self.coll.find({"createdAt": {"$gte": start, "$lte": end}})
